#!/usr/bin/perl -w
##########################################################################
# $Id: FaultDetector.pl,v 1.5 2004/03/22 03:38:45 gca Exp $
#
# Description: A simple Fault Detector
# - This utility runs as a non-terminating script, which monitors
#   processes for crash-faults.
# - This script is to be bundled with "repman.pl"
# - Processes to monitor are supplied in a file specified by the
#   user as an argument.  As the script runs, it periodically
#   reads the file and check to be sure all indicated processes
#   are still running.  If the script detects that a process has
#   terminated, it will re-start the process by invoking repman
#   with the arguments specified in the file
# - To shutdown FD & all monitored processes, just use <CTRL>-C
#
# - The format of the input file is as follows:
#       #comment
#       LABEL : REPMAN COMMANDS
#
# Use:	FaultDetector.pl [-h] [-v] -f <inputfile>
# Switches: -h : Display Help
#           -v : Use Verbose Mode to display activity
#           -f <file> : specifies user input file
#
# Limitations:
# - Labels between processes must be unique!
# - Assumes presence of a working "repman" in the path.
# - Is not thread safe! Assumes non-simultaneous execution by any given
#   user (note: could use flock, but this shouldn't really happen).
##########################################################################
use strict;

my $verbose = 0;		# for debugging...
my @MYARGV = @ARGV;             # argument list, saved for later...

# Default configuration values
my $user = $ENV{'USER'};         # the username under which commands are run
my $hostname = $ENV{'HOSTNAME'}; # the local hostname
my $repman = 'repman';	 # command to invoke ReplicationManager
my $rcFile = "$ENV{'HOME'}/.FDrc"; # the configuration file, stores %machineMap
my $delay  = 8;                  # delay between ping sessions

# Subroutine prototypes
sub FDHelp();          # Display usage
sub FDProcess($);      # read RC file & process labels/commands
sub FDLaunch(@);       # call repman to launch a process

sub isinarray(@);

sub trim;
#-------------------------------------------------------------------------
# Main - the main application logic
main: {
  print "\nStarting Fault Detector...\n";

  # process the first argument to determine the mode
  # - once mode determined, call the appropriate subroutine
  while(my $arg = shift) {
    # repman -k <label>
    if ($arg eq '-h') {        # Show Help/Usage
      &FDHelp();             # - Show help...
      exit;                  #   ...and terminate
    } elsif ($arg eq '-v') {   # Enable Verbose
      $verbose = 1;
    } elsif ($arg eq '-f') {
      $rcFile = shift or &FDHelp();
    }
  }

  # Begin monitoring process
  # - run in an endless loop, with a delay between interations
  # - Note: For FaultTolerance within FaultDetector do this:
  #   1) add initial FaultDetector process in RC file
  #   2) add a second FaultDetector process to RC file
  #   3) start initial FaultDetector via repman.
  #   Thus, FaultDetectors will monitor each other so that if
  #   one dies the other will restart it.
  if( $verbose ) {
    print "Verbose mode!\n" .
          "User: $user\@$hostname\n" .
          "Args[$#MYARGV]:@MYARGV\n" .
          "rcFile:$rcFile\n" .
          "delay:$delay\n";
  }
  while(1) {
    &FDProcess($rcFile);
    print "Sleeping for $delay seconds...\n";
    sleep $delay;            # sleep for n-seconds
  };
  print "Terminating...\n";
}

#-------------------------------------------------------------------------
# Displays usage syntax to user
sub FDHelp() {
  die "Syntax: FD.pl -h\n" .
      "        FD.pl [-v] -f <file>\n";
}

#-------------------------------------------------------------------------
# Processes the $rcFile
sub FDProcess($) {
  my $rcFile = shift or
    die "INTERNAL ERROR - Missing Argument!\n\t$!\n";
  my %processMap;

  # Open, Read & Close RC file quickly
  # - this permits updating file between processing sessions
  print "\nReading $rcFile...\n" if $verbose;
  open RC, $rcFile or die "ERROR! - Unable to open file: $rcFile\n\t$!";
  # Process each line of $rcFile
  LINE:  while(my $line = <RC>) {
    next LINE if $line eq "\n"; # skip blank lines
    next LINE if $line =~ /^#/; # skip comment lines
    chomp $line;                # remove trailing '\n' character
    print "Processing Line: $line\n" if $verbose;

    my ($label,$cmd) = split(/:/,$line,2);
    $cmd = &trim($cmd);
    $label = &trim($label);
    print "Label:-$label- \t Cmd:-$cmd-\n" if $verbose;
    $processMap{$label} = $cmd;

#    &ProcessLine($line);
  }
  close RC;

  # Invoke Repman to obtain list of running processes
  print "\nQuerying RepMan...\n" if $verbose;
  my $cmdline = "/bin/sh -c \"$repman\"";
  print " cmdline:$cmdline\n" if $verbose;
  my @repmanOutput = `$cmdline`; 
#    or die "ERROR! - system $cmdline failed!: $?\n\t$!";
  my %repmanMap;
  foreach my $line (@repmanOutput) {
    my ($label,$host) = split(" ",$line);
    $repmanMap{$label} = $host;
    print "Label:$label Host:$host\n" if $verbose;
  }
  my @labels = keys %repmanMap;

  # Processes map obtained from $rcFile
  # - Make sure process is still running.
  # - If it is not running, then start it using repman
  print "\nProcessing...\n";
  foreach my $label (keys %processMap) {
    my $cmd = $processMap{$label};
    print "Checking: $label ...";
    if( isinarray($label,@labels) ) {
      print "\tOk.\n";
    } else {
      print "\tRESTARTING!...\n";
      &FDLaunch($label,$cmd);
    }
  }

}

#-------------------------------------------------------------------------
# Launch a process via repman
sub FDLaunch(@) {
  my $label = shift or
    die "INTERNAL ERROR - Missing Label Argument!\n\t$!\n";
  my $cmd = shift or
    die "INTERNAL ERROR - Missing Cmd Argument!\n\t$!\n";

  print "Restarting '$label' with command '$cmd'\n";
  my $cmdline = "$repman $cmd";
  print " $cmdline\n" if $verbose;
  system($cmdline) == 0
    or die "ERROR! - system $cmdline failed $?\n\t$!\n";
}

#-------------------------------------------------------------------------
# - simple function to determine if the first argument (string) exists
#   in the second argument (array of strings).
# Return: A simple TRUE/FALSE
sub isinarray(@) {
  my $value = shift;
  my @array = @_;
ELEM: foreach my $elem (@array) {
    if( $value eq $elem ) {
      return 1;
      last;
    }
  }
  return 0;
}

#-------------------------------------------------------------------------
# trims leading & trailing whitespace
# use:
#   $string = trim($string);
#   @many   = trim(@many);
# see: http://iis1.cps.unizar.es/Oreilly/perl/cookbook/ch01_15.htm

sub trim {
    my @out = @_;
    for (@out) {
        s/^\s+//;
        s/\s+$//;
    }
    return wantarray ? @out : $out[0];
}

