package com.spamnbeans.client;

import java.io.*;
import java.util.*;
import java.util.logging.*;

import com.spamnbeans.msg.MsgStats;

/**
 * Tracks statistics for a given message, as well as global statistics.
 */
public class ClientStats {
    private static Logger logger = Logger.getLogger(EatSpam.class.getName());

	private static List statList = Collections.synchronizedList(new LinkedList());
	private static int messageCount = 0;
	private static int errorCount = 0;
	private static long totalTime = 0;
	private static long totalBytes = 0;

	private MsgStats msgStats = null;
	private long startTime = 0;
	private long elapsedTime = 0;
	private int size = 0;

	public void setSize(int i) {
		size = i;
	}

	public void start() {
		startTime = System.currentTimeMillis();
	}

	private void end() {
		if (startTime == 0) {
			throw new NullPointerException("start must be called before end");
		}

		long endTime = System.currentTimeMillis();
		elapsedTime = endTime - startTime;
	}

	/**
	 * End without an error.
	 */
	public void end(MsgStats msgStats) {
		end();
		msgStats.setTotalTime(elapsedTime);
		this.msgStats = msgStats;
		tabulate(this);
	}

	/**
	 * End, with an error.
	 */
	public void end(Throwable t) {
		// NOTE ignore error for now
		end();
		tabulate(this);
	}

	private static synchronized void tabulate(ClientStats stats) {
		if (stats.msgStats == null) {
			errorCount++;
		}
		else {
			statList.add(stats.msgStats);
		}
		messageCount++;
		totalTime += stats.elapsedTime;
		totalBytes += stats.size;
	}

	public static void logResults() {
		long elapsedSeconds = totalTime / 1000;
		logger.info("Messages processed: "+messageCount);
		logger.info("   error count: "+errorCount);
		logger.info("Elapsed seconds: "+elapsedSeconds);

		if (messageCount > 0) {
			long averageSeconds = 100 * elapsedSeconds / messageCount;
			logger.info("   seconds/message: "+((float)averageSeconds/100));

			if (elapsedSeconds > 0) {
				long messageRate = 100 * messageCount / elapsedSeconds;
				logger.info("   messages/second: "+((float)messageRate/100));
			}

			logger.info("Byes processed: "+totalBytes);

			long averageBytes = 100 * totalBytes / messageCount;
			logger.info("   bytes/message: "+((float)averageBytes/100));

			if (elapsedSeconds > 0) {
				long byteRate = 100 * totalBytes / elapsedSeconds;
				logger.info("   bytes/second: "+((float)byteRate/100));
			}
		}
		graphStats();
	}

	private static void graphStats() {
		try {
			String[] types = new String[] {
				"Client", "Server", "Spam", "DB"
			};
			PrintStream[] ps = new PrintStream[types.length];
			for (int i = 0; i < types.length; i++) {
				ps[i] = new PrintStream(new FileOutputStream(
					new File(System.getProperty("java.io.tmpdir"), types[i]+"Stats.dat")
				));
			}

			for (Iterator i = statList.iterator(); i.hasNext(); ) {
				MsgStats msgStats = (MsgStats)i.next();
				ps[0].println(msgStats.getTotalTime() - msgStats.getServerTime());
				ps[1].println(msgStats.getServerTime() - msgStats.getSpamTime() - msgStats.getDbTime());
				ps[2].println(msgStats.getSpamTime());
				ps[3].println(msgStats.getDbTime());
			}

			for (int i = 0; i < types.length; i++) {
				ps[i].close();
			}
		}
		catch (IOException e) {
			if (logger.isLoggable(Level.FINE)) {
				logger.log(Level.FINE, "unable to save stats graph", e);
			}
			else {
				logger.warning("unable to save stats graph:"+e);
			}
		}
	}

}
