package com.spamnbeans.client;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.logging.*;
import java.util.regex.*;

import com.spamnbeans.msg.MsgStats;

/**
 * A pseudo-MTA, responsible for receiving messages (currently off of disk),
 * handing them to the MPClient for processing, and then delivering them
 * (currently by reporting statistics on them).
 */
public class EatSpam implements Runnable {
    private static Logger logger = Logger.getLogger(EatSpam.class.getName());

	private static final String baseDir = "/afs/ece/usr/cfry/classes/middleware";
	private static final String publicCorpus = baseDir + "/publiccorpus";
	private static final String assassinDir = baseDir + "/spamassassin/usr/bin";
	private static final String javaCommand = "java -Djava.util.logging.config.file="+ baseDir +"/config/logging.properties";

	private static final boolean sendEasyHam = true;
	private static final boolean sendHardHam = true;
	private static final boolean sendSpam = true;
	private static final boolean mixCategories = true; // as opposed to 1 at a time

	private static final int DEFAULT_MESSAGE_COUNT = 10;

	private int messageCount;
	private MPClient client = null;
	private List list = null;

	// NOTE: having EatSpam objects instead of putting everything in main will
	// facilitate the transition to multiple threads

	public EatSpam() throws Throwable {
		this(DEFAULT_MESSAGE_COUNT);
	}

	public EatSpam(int messageCount) throws Throwable {
		this(messageCount, new Random());
	}

	public EatSpam(int messageCount, long seed) throws Throwable {
		this(messageCount, new Random(seed));
	}

	public EatSpam(int messageCount, Random random) throws Throwable {
		this.messageCount = messageCount;
		try {
			client = new MPClient();
		}
		catch (Throwable t) {
			logDeath("Error creating client", t);
			throw t;
		}

		try {
			list = getMessageList();
			Collections.shuffle(list, random);
		}
		catch (Throwable t) {
			logDeath("Error reading message lists", t);
			throw t;
		}
	}

	public void run() {
		int count = 0;
		for (Iterator i = list.iterator(); i.hasNext() && count < messageCount; count++) {
			ClientStats clientStats = new ClientStats();
			try {
				String file = (String)i.next();
				String[] s = client.readMessage(
					new FileInputStream(publicCorpus + "/" + file)
				);

				clientStats.setSize(s[0].length() + s[1].length());
				clientStats.start();
				MsgStats msgStats = client.processMessage(s[0] + s[1]);
				clientStats.end(msgStats);

				String newHeaders = msgStats.getNewHeaders();
				// TODO missing retry statictics

				if (logger.isLoggable(Level.INFO)) {
					String[] lines = newHeaders.split("\n");
					String levelLine = lines[lines.length-1];
					Pattern p = Pattern.compile("X-Spam-Level: ?(\\**)$");
					Matcher m = p.matcher(levelLine);
					int spamLevel;
					if (m.matches()) {
						String stars = m.group(1);
						spamLevel = stars.length();
					}
					else {
						logger.severe("Couldn't determine spam level: "+levelLine);
						spamLevel = 0;
					}

					if (file.startsWith("easy_ham")) {
						logger.info(
							"EasyHam, marked with spam level " + spamLevel
						);
					}
					else if (file.startsWith("hard_ham")) {
						logger.info(
							"HardHam, marked with spam level " + spamLevel
						);
					}
					else if (file.startsWith("spam")) {
						logger.info(
							"Spam, marked with spam level " + spamLevel
						);
					}
					else {
						logger.info(
							"Unknown source ("+file+"), marked with spam level " + spamLevel
						);
					}

					if (logger.isLoggable(Level.FINE)) {
						// print out detailed message stats
						logger.fine("  totalTime  = "+msgStats.getTotalTime()+" ms");
						logger.fine("    serverTime = "+msgStats.getServerTime()+" ms");
						logger.fine("      spamTime = "+msgStats.getSpamTime()+" ms");
						logger.fine("      dbTime   = "+msgStats.getDbTime()+" ms");
					}
				}
			}
			catch (Throwable t) {
				if (logger.isLoggable(Level.FINER)) {
					logger.throwing(EatSpam.class.getName(), "run", t);
				}
				else {
					logger.warning("Error encountered while processing message: "+t);
				}
				clientStats.end(t);
				// TODO handle fatal/non-fatal
			}
		}

		ClientStats.logResults();
	}

	private static List readMessageList(String file) throws IOException {
		List list = new ArrayList(4000);
		BufferedReader in = new BufferedReader(new FileReader(file));
		String inputLine;
		while ((inputLine = in.readLine()) != null) {
			list.add(inputLine);
		}
		return list;
	}

	private static List getMessageList() throws IOException {
		List list = new ArrayList(7000);
		if (sendEasyHam) {
			list.addAll(readMessageList(publicCorpus + "/easy_ham.list"));
		}
		if (sendHardHam) {
			list.addAll(readMessageList(publicCorpus + "/hard_ham.list"));
		}
		if (sendSpam) {
			list.addAll(readMessageList(publicCorpus + "/spam.list"));
		}
		return list;
	}

	private static void printResults() {
	}

	private static void logDeath(String message, Throwable t) {
		if (logger.isLoggable(Level.FINEST)) {
			logger.log(Level.FINEST, message, t);
		}
		else {
			if (t != null) {
				message = message + t;
			}
			logger.severe(message);
		}
		ClientStats.logResults();
	}

	public static void main(String[] args) throws Throwable {
		EatSpam eatSpam = new EatSpam();
		if (args.length >= 1) {
			int messageCount = Integer.parseInt(args[0]);
			// first arg, if specified, is message count
			if (args.length >= 2) {
				long seed = Long.parseLong(args[1]);
				eatSpam = new EatSpam(messageCount, seed);
			}
			else {
				eatSpam = new EatSpam(messageCount);
			}
		}
		else {
			eatSpam = new EatSpam();
		}
		try {
			new Thread(eatSpam).start();
		}
		catch (Throwable t) {
			// this is only for errors in the constructor
			// all such errors have already been logged, we just need to
			// exit with a failure code (once multiple threads are added,
			// a different response may be appropriate)
			System.exit(5342); // only exit because we are top-level client
		}
	}

}
