// $Id: ManagerBean.java,v 1.12 2004/03/23 00:01:12 gca Exp $

package com.spamnbeans.server;

import com.spamnbeans.server.*;
import com.spamnbeans.msg.*;

import java.rmi.RemoteException;
import javax.rmi.PortableRemoteObject;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.ejb.SessionBean;
import javax.ejb.EJBException;
import javax.ejb.CreateException;
import javax.ejb.FinderException;
import javax.ejb.RemoveException;
import java.util.*;
import java.util.logging.*;
import java.io.PrintWriter;
import java.io.StringWriter;

/**
 * Define the ManagerBean methods.
 * 
 * <p>
 * The ManagerBean is responsible for creating, finding, and removing
 * ClientBeans and ServerBeans, which are used mainly for validation and
 * statistics gathering.
 * 
 * @author Andrew Boyer, <a href="mailto:aboyer@andrew.cmu.edu">
 *         aboyer@andrew.cmu.edu</a>
 * @version 1.0
 */
public class ManagerBean implements SessionBean {
	private ServerStatHomeLocal sshl;
	private ClientStatHomeLocal cshl;
	private MsgHomeLocal mhl;

	private static Logger logger =
		Logger.getLogger(ManagerBean.class.getName());

	/**
	 * Create a new ManagerBean Gets and stores initial HomeLocal interfaces
	 * for Servers and Clients
	 */
	public void ejbCreate() throws RemoteException {
		try {
			Context jndiContext = new InitialContext();

			Object sobj = jndiContext.lookup("ServerStatHomeLocal");
			sshl =
				(ServerStatHomeLocal) PortableRemoteObject.narrow(
					sobj,
					ServerStatHomeLocal.class);

			Object cobj = jndiContext.lookup("ClientStatHomeLocal");
			cshl =
				(ClientStatHomeLocal) PortableRemoteObject.narrow(
					cobj,
					ClientStatHomeLocal.class);

			Object mobj = jndiContext.lookup("MsgHomeLocal");
			mhl =
				(MsgHomeLocal) PortableRemoteObject.narrow(
					mobj,
					MsgHomeLocal.class);
		} catch (Exception e) {
			throw new EJBException(e);
		}
	}

	/**
	 * Create a new Server EntityBean
	 * 
	 * @param serverID
	 *            The ServerID (primary key) of the created Bean
	 */
	public void serverCreate(Integer serverID, String name, boolean isEnabled)
		throws RemoteException, CreateException, DuplicateServerException {
		try {
			ServerStatLocal ssl = sshl.findByPrimaryKey(serverID);
			throw new DuplicateServerException(
				"ServerID " + serverID + " already exists");
		} catch (FinderException fe) {
			// If server not found - ok to continue
		}
		try {
			ServerStatLocal ssl = sshl.findByName(name);
			throw new DuplicateServerException(
				"Server " + name + " already exists");
		} catch (FinderException fe) {
		}

		ServerStatLocal ssl = sshl.create(serverID, name, isEnabled);
	}

	/**
	 * Remove Server EntityBean
	 * 
	 * @param serverID
	 *            The ServerID of the removed Bean
	 */
	public void serverRemove(Integer serverID)
		throws RemoteException, InvalidServerException, RemoveException {
		try {
			ServerStatLocal ssl = sshl.findByPrimaryKey(serverID);
			ssl.remove();
		} catch (FinderException fe) {
			throw new InvalidServerException(
				"ServerID " + serverID + " does not exist");
		}
	}

	/**
	 * Check if a Server EntityBean already exists
	 * 
	 * @param serverID
	 *            The ServerID of the Bean to check
	 */
	public boolean serverExists(Integer serverID) throws RemoteException {
		try {
			ServerStatLocal ssl = sshl.findByPrimaryKey(serverID);
			return true;
		} catch (Throwable t) {
			return false;
		}
	}

	/**
	 * Get the Name of a Server EntityBean
	 * 
	 * @param serverID
	 *            The ServerID of the Bean to check
	 */
	public String serverGetName(Integer serverID)
		throws RemoteException, InvalidServerException {
		try {
			ServerStatLocal ssl = sshl.findByPrimaryKey(serverID);
			return ssl.getName();
		} catch (FinderException fe) {
			throw new InvalidServerException(
				"ServerID " + serverID + " does not exist");
		}
	}

	/**
	 * Get the ServerID of a Server EntityBean
	 * 
	 * @param name
	 *            The Name of the Bean to check
	 */
	public Integer serverGetID(String name)
		throws RemoteException, InvalidServerException {
		try {
			ServerStatLocal ssl = sshl.findByName(name);
			return ssl.getServerID();
		} catch (FinderException fe) {
			throw new InvalidServerException(
				"Server " + name + " does not exist");
		}
	}

	/**
	 * Get a list of all existing Server EntityBeans by ID
	 */
	public HashSet serverListIDs() throws RemoteException, FinderException {
		HashSet ids = new HashSet();
		Collection servers = sshl.findAll();
		Iterator it = servers.iterator();

		while (it.hasNext()) {
			ServerStatLocal ssl = (ServerStatLocal) it.next();
			ids.add(ssl.getServerID());
		}
		return ids;
	}

	/**
	 * Get a list of all existing Server EntityBeans by name
	 */
	public HashSet serverListNames() throws RemoteException, FinderException {
		HashSet ids = new HashSet();
		Collection servers = sshl.findAll();
		Iterator it = servers.iterator();

		while (it.hasNext()) {
			ServerStatLocal ssl = (ServerStatLocal) it.next();
			ids.add(ssl.getName());
		}
		return ids;
	}

	/**
	 * Check if a Server EntityBean is active
	 * 
	 * @param name
	 *            The name of the Bean to check
	 */
	public boolean serverIsEnabled(String name)
		throws RemoteException, InvalidServerException {
		try {
			ServerStatLocal ssl = sshl.findByName(name);
			boolean isenabled = ssl.getEnable();
			return isenabled;
		} catch (FinderException fe) {
			throw new InvalidServerException("Server " + name + " not found");
		}
	}
	public boolean serverIsEnabled(Integer serverID)
		throws RemoteException, InvalidServerException {
		try {
			ServerStatLocal ssl = sshl.findByPrimaryKey(serverID);
			return ssl.getEnable();
		} catch (FinderException fe) {
			throw new InvalidServerException(
				"Server " + serverID + " not found");
		}
	}

	/**
	 * Set a Server EntityBean's activity level
	 * 
	 * @param name
	 *            The name of the Bean to change
	 */
	public void serverSetEnabled(String name, boolean Enable)
		throws RemoteException, InvalidServerException {
		try {
			ServerStatLocal ssl = sshl.findByName(name);
			Integer serverID = ssl.getServerID();
			ssl.setEnable(Enable);
			String strMsg = "ManagerBean:serverSetEnabled["
				+ name + ","
				+ serverID + ","
				+ Enable + "]";
			logger.info(strMsg);
		} catch (FinderException fe) {
			throw new InvalidServerException("Server " + name + " not found");
		}
	}
	public void serverSetEnabled(Integer serverID, boolean Enable)
		throws RemoteException, InvalidServerException {
		try {
			ServerStatLocal ssl = sshl.findByPrimaryKey(serverID);
			String name = ssl.getName();
			ssl.setEnable(Enable);
			String strMsg = "ManagerBean:serverSetEnabled["
				+ name + ","
				+ serverID + ","
				+ Enable + "]";
			logger.info(strMsg);
		} catch (FinderException fe) {
			throw new InvalidServerException(
				"Server " + serverID + " not found");
		}
	}

	/**
	 * Get a list of all Enabled Server EntityBeans
	 */
	public HashSet serverListEnabled()
		throws RemoteException, FinderException {
		HashSet names = new HashSet();
		Collection servers = sshl.findEnabled(true);
		Iterator it = servers.iterator();

		while (it.hasNext()) {
			ServerStatLocal ssl = (ServerStatLocal) it.next();
			names.add(ssl.getName());
		}

		return names;
	}

	/**
	 * Get the name of one active Server EntityBean
	 */
	public String serverLoadBalance()
		throws RemoteException, InvalidServerException {
		try {
			// TODO Contact ReplicationManager for Replica
			Collection servers = sshl.findEnabled(true);
			Iterator it = servers.iterator();

			Random rgen = new Random();
			ServerStatLocal ssl = null;

			int rint = rgen.nextInt(servers.size()) + 1;
			System.out.println("Load Balancer: " + rint + " " + servers.size());
			for (int i = 0; i < rint; i++)
				ssl = (ServerStatLocal) it.next();

			return ssl.getName();
		} catch (Throwable t) {
			throw new InvalidServerException("No server active");
		}
	}

	/**
	 * Create a new Client EntityBean
	 * 
	 * @param clientID
	 *            The ClientID (primary key) of the created Bean
	 */
	public void clientCreate(Integer clientID, String name)
		throws RemoteException, CreateException, DuplicateClientException {
		try {
			ClientStatLocal csl = cshl.findByPrimaryKey(clientID);
			throw new DuplicateClientException(
				"ClientID " + clientID + " already exists");
		} catch (FinderException fe) {
		}

		try {
			ClientStatLocal csl = cshl.findByName(name);
			throw new DuplicateClientException(
				"Client " + name + " already exists");
		} catch (FinderException fe) {
		}

		ClientStatLocal csl = cshl.create(clientID);
		csl.setName(name);
	}

	/**
	 * Remove Client EntityBean
	 * 
	 * @param clientID
	 *            The ClientID of the removed Bean
	 */
	public void clientRemove(Integer clientID)
		throws RemoteException, InvalidClientException, RemoveException {
		try {
			ClientStatLocal csl = cshl.findByPrimaryKey(clientID);
			csl.remove();
		} catch (FinderException fe) {
			throw new InvalidClientException(
				"ClientID " + clientID + " does not exist");
		}
	}

	/**
	 * Check if a Client EntityBean already exists
	 * 
	 * @param clientID
	 *            The ClientID of the Bean to check
	 */
	public boolean clientExists(Integer clientID) throws RemoteException {
		try {
			ClientStatLocal csl = cshl.findByPrimaryKey(clientID);
			return true;
		} catch (Throwable t) {
			return false;
		}
	}

	/**
	 * Set the Name of a Client EntityBean
	 * 
	 * @param clientID
	 *            The ClientID of the Bean to set
	 * @param name
	 *            The Name to set
	 */
	public void clientSetName(Integer clientID, String name)
		throws RemoteException, InvalidClientException, DuplicateClientException {
		try {
			ClientStatLocal csl = cshl.findByName(name);
			throw new DuplicateClientException(
				"Client " + name + " already exists");
		} catch (FinderException fe) {
		} // Ignore

		try {
			ClientStatLocal csl = cshl.findByPrimaryKey(clientID);
			csl.setName(name);
		} catch (FinderException fe) {
			throw new InvalidClientException(
				"ClientID " + clientID + " does not exist");
		}
	}

	/**
	 * Get the Name of a Client EntityBean
	 * 
	 * @param clientID
	 *            The ClientID of the Bean to check
	 */
	public String clientGetName(Integer clientID)
		throws RemoteException, InvalidClientException {
		try {
			ClientStatLocal csl = cshl.findByPrimaryKey(clientID);
			return csl.getName();
		} catch (FinderException fe) {
			throw new InvalidClientException(
				"ClientID " + clientID + " does not exist");
		}
	}

	/**
	 * Get the ClientID of a Client EntityBean
	 * 
	 * @param name
	 *            The Name of the Bean to check
	 */
	public Integer clientGetID(String name)
		throws RemoteException, InvalidClientException {
		try {
			ClientStatLocal csl = cshl.findByName(name);
			return csl.getClientID();
		} catch (FinderException fe) {
			throw new InvalidClientException(
				"Client " + name + " does not exist");
		}
	}

	/**
	 * Get a list of all running Client EntityBeans
	 */
	public HashSet clientListIDs() throws RemoteException, FinderException {
		HashSet ids = new HashSet();
		Collection clients = cshl.findAll();
		Iterator it = clients.iterator();

		while (it.hasNext()) {
			ClientStatLocal csl = (ClientStatLocal) it.next();
			ids.add(csl.getClientID());
		}
		return ids;
	}

	/**
	 * Get stats of Msg EntityBeans for a particular Client EntityBean TODO:
	 * Update to return a class/struct with all stat info
	 */
	public String msgGetStatsByClient(Integer clientID)
		throws RemoteException, FinderException {
		int totalMsgCount = 0;
		long totalMsgBytes = 0;
		String output = null;
		StringWriter writer = new StringWriter();
		PrintWriter out = new PrintWriter(writer);

		// Make sure clientID is valid
		if (!clientExists(clientID)) {
			out.println("Client " + clientID + " does not exist");
		}

		// Scan Msg beans & accumulate total stats
		try {
			Collection msgs = mhl.findByClientID(clientID);
			Iterator it = msgs.iterator();
			while (it.hasNext()) {
				MsgLocal msgl = (MsgLocal) it.next();
				totalMsgCount += 1;
				totalMsgBytes += msgl.getMsgSize().longValue();
			};
		} catch (FinderException fe) {
			// No records exist for this client
		}
		out.println("#:" + totalMsgCount + " Bytes:" + totalMsgBytes);
		output = writer.toString();
		return output;
	}

	/**
	 * Remove Msg EntityBean
	 * 
	 * @param msgID
	 *            The MsgID of the removed Bean
	 */
	public void msgRemove(Integer msgID)
		throws RemoteException, InvalidMessageIDException, RemoveException {
		try {
			MsgLocal ml = mhl.findByPrimaryKey(msgID);
			ml.remove();
		} catch (FinderException fe) {
			throw new InvalidMessageIDException(
				"MsgID " + msgID + " does not exist");
		}
	}

	/**
	 * Get a list of all Msg EntityBeans
	 */
	public HashSet msgListIDs() throws RemoteException, FinderException {
		HashSet ids = new HashSet();
		Collection msgs = mhl.findAllMsgs();
		Iterator it = msgs.iterator();

		while (it.hasNext()) {
			MsgLocal ml = (MsgLocal) it.next();
			ids.add(ml.getMsgID());
		}
		return ids;
	}

	/**
	 * Not implemented
	 */
	public void ejbRemove() {
	}

	/**
	 * Not implemented
	 */
	public void ejbActivate() {
	}

	/**
	 * Not implemented
	 */
	public void ejbPassivate() {
	}

	/**
	 * Not implemented
	 */
	public void setSessionContext(javax.ejb.SessionContext cntx) {
	}
}
